#!/usr/bin/env python
# coding: utf-8

"""
Convenience API for the ini file

Created: 09.05.20
"""

import os
import json
import configparser
from pathlib import Path

_CONFIG = None

def read():
    global _CONFIG
    if not _CONFIG:
        config_folder = Path(os.environ["HOME"]) / ".luechenbresse"
        ini_file = config_folder / "luechenbresse.ini"
        _CONFIG = configparser.ConfigParser()
        _CONFIG.read(ini_file)

def get(section, key, default=None):
    read()
    if section in _CONFIG:
        if key in _CONFIG[section]:
            return _CONFIG[section][key]
    return default

if __name__ == "__main__":
    pass


#!/usr/bin/env python
# coding: utf-8

"""
Tooling to deal with nested lists of strings (e.g. extracted from a web page),
transform them to a list of words without decorations and then work with such list.
"""

_DECORATORS = '''+"'-.!?;:,|/“”()'''

def undecorate(mixup):
    # get undecorated words
    # does not change the nesting, but adds a level everywhere
    # mixup: str | list(str) | nested mixups -> nested mixup one level deeper
    if isinstance(mixup, list) or isinstance(mixup, tuple):
        l = [undecorate(s) for s in mixup]
    else:
        assert isinstance(mixup, str), (mixup, type(mixup))
        l = [word.strip(_DECORATORS) for word in mixup.split(" ") if word != ""]
    return [e for e in l if e != ""]

def flatten(mixup):
    # make nested list of strings a flat list
    # does not change the strings, but flattens the list structure only
    # mixup: nested mixup of lists of str -> list(str)
    if isinstance(mixup, str):
        return mixup
    else:
        assert isinstance(mixup, list) or isinstance(mixup, tuple), (mixup, type(mixup))
        l = list()
        for y in mixup:
            f = flatten(y)
            if isinstance(f, str):
                l.append(f)
            else:
                l.extend(flatten(y))
        return l

def join(*args):
    # form one flat list of undecorated words from nested list(s) of strings
    l = list()
    for arg in args:
        l.extend(flatten(undecorate(arg)))
    return l

def sjoin(*args):
    # form one string of undecorated words from nested list(s) of strings
    return " ".join(join(args))  # args is tuple, not list

def context(words, i, j, width=5):
    """
    words: list(str)
    i, j, width: int, all in terms of words not chars
    indexes i..j in l are the string, add context of with <width> words both sides
    returns the context string plus the index of the first letter of l[i]
    [ 0 - - - - - - i - - j - - - - - - ], width=3
              - - - i - - j - - -
    context is surrounded by dots when l has more
    """
    if j < i:
        i, j = j, i
    if i < 0:
        i = 0
    i0 = max(i-width, 0)
    j0 = min(j + width + 1, len(words))
    with_context = " ".join(words[i0:j0])
    offset = len(" ".join(words[i0:i]))
    if i > 0 and offset > 0:
        offset += 1
    if i0 > 0:
        with_context = "... " + with_context
        offset += 4
    if j0 < len(words):
        with_context = with_context + " ..."
    assert with_context[offset] == words[i][0]
    return with_context, offset

def print_aligned_context(ctxs):
    # ctxs: list(tuple), tuples like returnes by context()
    # prints like so:
    #                                indent |
    #                                        Nordkorea Kim kündigt neue Waffe ...
    #                     ... wird sich laut Machthaber Kim nicht mehr an ...
    #             ... strategischen Waffe an Nordkoreas Machthaber Kim Jong Un ...
    #                ... Waffe an Nordkoreas Machthaber Kim Jong Un hat ...
    indent = max([x[1] for x in ctxs])
    for ctx in ctxs:
        print(" " * (indent - ctx[1]), ctx[0])


if __name__ == "__main__":
    pass



#!/usr/bin/env python
# coding: utf-8

"""
Convenience API to Mailgun

Created: 11.05.20
"""

import requests
import logging

from luechenbresse import ini


# courtesy https://gist.github.com/lalzada/3938daf1470a3b7ed7d167976a329638
class Mailgun(object):

    def __new__(cls):
        if not hasattr(cls, 'instance') or not cls.instance:
            cls.instance = super().__new__(cls)
            mailgun = "mailgun"
            cls.instance.url = ini.get(mailgun, "url")
            cls.instance.auth_key = ini.get(mailgun, "auth-key")
            cls.instance.mail_from = ini.get(mailgun, "from")
            cls.instance.mail_to = ini.get(mailgun, "to")
            cls.instance.active = cls.instance.url and cls.instance.auth_key and cls.instance.mail_from and cls.instance.mail_to
        return cls.instance

    def shoot(self, subject, body):
        if not self.active:
            logging.info("no mailgun account configured")
        else:
            logging.info(f"sending mail: {subject}")
            try:
                r = requests.post(
                    self.url,
                    auth=("api", self.auth_key),
                    data={
                        "from": self.mail_from,
                        "to": self.mail_to,
                        "subject": subject,
                        "text": body
                    })
                logging.info(f"HTTP {r.status_code}")
            except Exception as ex:
                logging.exception(f"{ex.__class__.__name__}: {str(ex)}")

if __name__ == "__main__":
    pass


#!/usr/bin/env python
# coding: utf-8

"""
Special coding for the zdf-heute RSS feed.
We can re-think the exit-mechanism later (class/interface, whatever...).

Created: 09.05.20
"""

from datetime import datetime
from time import mktime

def parse_feed_header(feed_header):
    return feed_header["title"], feed_header["published"]

def parse_feed_entry(feed_entry):
    url = feed_entry["link"]
    rss_id = feed_entry["id"]
    title = feed_entry["title"]
    ts = datetime.fromtimestamp(mktime(feed_entry["published_parsed"])).isoformat()
    return url, rss_id, title, ts


if __name__ == "__main__":
    pass


#!/usr/bin/env python
# coding: utf-8

"""
Cares for downloading the feed and the references articles.

Created: 09.05.20
"""

import json
from time import time, mktime, sleep
from datetime import datetime
import random
import sqlite3
from pathlib import Path
import importlib
import logging

import requests
import feedparser

from luechenbresse import ini
from luechenbresse import data


# mimicks a very simple Request object when a request failed
class FakeRequest:
    def __init__(self):
        self.status_code = 999
        self.text = "Hurz."


# tame network requests
def hold_on():
    bubu = random.uniform(3, 8)
    logging.info(f"sleeping {bubu:0.2f} sec")
    sleep(bubu)


class Feed:

    @staticmethod
    def from_json(name, feed):
        return Feed(name, feed["feed"], feed["type"], feed["db"], feed["type"])

    @staticmethod
    def from_name(name):
        feed_list = data.feeds()
        feed = feed_list[name]
        return Feed.from_json(name, feed)

    def __init__(self, name, feed, type, db, schema):
        """
        Constructor.
        :param feed: url of the feed
        :param type: rss|...
        :param db: name of the db file (location from luechenbresse.ini)
        :param schema: list(str) of schemas to apply
        """
        logging.info(f"Feed.__init__: {name} ({type}) from {feed}")
        self.name = name
        self.feed = feed
        self.type = type
        ini_path = ini.get("databases", "folder")
        if not ini_path:
            raise ValueError(f"No database folder specified in luechenbresse.ini")
        ini_file = Path(ini_path)
        self.db = ini_file / db
        self.schema = schema # TODO support more than one schema
        feed_module_name = "luechenbresse."+name
        logging.info(f"loading module {feed_module_name}")
        self.feed_module = importlib.import_module(feed_module_name)
        # methods will open and close connection if not done from outside
        self.conn = None
        self.cur = None

    def _open_db(self):
        self.conn = sqlite3.connect(self.db)
        self.cur = self.conn.cursor()

    def _close_db(self):
        self.conn.close()
        self.conn = None
        self.cur = None

    def _is_there(self, url):
        sql = """
            SELECT * 
            FROM articles 
            WHERE url = ?
        """
        rows = [r for r in self.cur.execute(sql, (url,))]
        return len(rows) == 1

    def _upsert_article(self, url, rss_id, title, ts):
        sql = """
            INSERT INTO articles(url, rss_id, title, ts, realised_ts) 
                VALUES (?, ?, ?, ?, ?)
                ON CONFLICT DO NOTHING
        """
        now = datetime.now().isoformat()[:19]
        row_tbi = (url, rss_id, title, ts, now)
        self.cur.execute(sql, row_tbi)

    def _cleansed_entries(self, fp_feed):
        # remove duplicates
        unique = dict()
        cnt = 0
        cnt_no_url = 0
        for raw_entry in fp_feed["entries"]:
            cnt += 1
            entry = self.feed_module.parse_feed_entry(raw_entry)
            if entry[0]: # entry[0]=url         # TODO use named tuple
                unique[entry[0]] = entry
            else:
                cnt_no_url += 1
                logging.info(f"empty URL for '{entry[2]}' dated '{entry[3]}'") # 2=title, 3=ts
        # sort by time
        # TODO list comprehension
        entries = list()
        for url in unique:
            entries.append(unique[url])
        entries.sort(key=lambda tup: tup[3], reverse=True)  # tup[3]=ts
        logging.info(f"{cnt} entries, {cnt_no_url} w/o URL, duplicates removed, remaining: {len(entries)} ")
        return entries

    def get_rss(self):
        logging.info(f"GET {self.name}")
        t0 = time()
        # TODO catch exceptions
        f = feedparser.parse(self.feed)
        #dt = f"{time() - t0:0.3f}s"
        title, published = self.feed_module.parse_feed_header(f["feed"])
        logging.info(f"feed: {title}")
        if published:
            logging.info(f"published: {published}")
        logging.info(f"got from the internet in {time() - t0:0.3f} sec")
        private_connection = False
        if not self.cur:
            private_connection = True
            self._open_db()
        # sort entries by their own timestamps
        entries = self._cleansed_entries(f)
        cnt_new = 0
        cnt_old = 0
        for entry in entries:
            url, rss_id, title, ts = entry
            if self._is_there(url):
                cnt_old += 1
            else:
                cnt_new += 1
                logging.info(f"{ts} {title}")
                self._upsert_article(url, rss_id, title, ts)
        logging.info(f"{cnt_new} new, {cnt_old} known")
        self.conn.commit()
        if private_connection:
            self._close_db()
        logging.info(f"get_rss() total: {time() - t0:0.3f}s")

    def _get_backlog(self):
        private_connection = False
        if not self.cur:
            private_connection = True
            self._open_db()
        # avoid code for removal of the empty URL in the database, can still be done mmanually
        self.cur.execute("""
            SELECT url, ts, title FROM articles
            WHERE ( dl_http != 200 OR dl_http IS NULL ) AND url != ''
            ORDER BY ts
        """)
        rows = self.cur.fetchall()
        if private_connection:
            self._close_db()
        a = list()
        for row in rows:
            a.append({
                "url": row[0],
                "ts": row[1],
                "title": row[2]
            })
        logging.info(f"Backlog: {len(a)} Artikel")
        return a

    def _download_article(self, a):
        # a like returned by get_backlog()
        logging.info(f'downloading {a["ts"]} –– {a["title"]}')
        t0 = time()
        try:
            r = requests.get(a["url"])
        except Exception as ex:
            logging.error(ex)
            logging.debug("faking request object")
            r = FakeRequest()
        dt = time() - t0
        logging.info(f'{a["url"]}: {dt:0.3f}s HTTP {r.status_code}')
        text = r.text if r.status_code == 200 else None
        now = datetime.now().isoformat()[:19]
        row = (now, r.status_code, dt, text, a["url"])
        private_connection = False
        if not self.cur:
            private_connection = True
            self._open_db()
        self.cur.execute("""
            UPDATE articles
            SET dl_ts = ?, dl_http = ?, dl_dt = ?, html = ?
            WHERE url = ?
        """, row)
        self.conn.commit()
        if private_connection:
            self._close_db()

    def process_backlog(self):
        logging.info(f"Processing backlog for {self.name}")
        backlog = self._get_backlog()
        for i, article in enumerate(backlog):
            logging.info(i)
            self._download_article(article)
            hold_on()


def process_feed(name): # TODO make @staticmethod of Feed
    feed = Feed.from_name(name)
    feed.get_rss()
    feed.process_backlog()

def process_all_feeds(): # TODO make @staticmethod of Feed
    logging.debug("process_all_feeds()")
    feeds = data.feeds()
    logging.debug(f"{len(feeds)} feeds")
    for name in feeds:
        feed = Feed.from_name(name)
        feed.get_rss()
    for name in feeds:
        feed = Feed.from_name(name)
        feed.process_backlog()

if __name__ == "__main__":
    pass


#!/usr/bin/env python
# coding: utf-8

"""
Collect code that works with the files in teh .luechenbresse folder directly.

Created: 15.05.20
"""

import sys
import os
from pathlib import Path
import sqlite3
import logging
import configparser

from luechenbresse import data
from luechenbresse import __version__ as luechenbresse_version      # TODO gefällt mir nicht
from luechenbresse.mailgun import Mailgun

def _create_db(db_file, schema):
    # TODO support more than one schema per db
    if isinstance(schema, list):
        schema = schema[0]
    sql, _ = data.schema(schema)
    con = sqlite3.connect(db_file)
    cur = con.cursor()
    cur.executescript(sql)
    con.close()

def init_dotfolder(db_folder):
    # TODO move init() to module

    # create ~/.luechenbresse
    config_folder = Path(os.environ["HOME"]) / ".luechenbresse"
    logging.info(f"Configuration folder: {config_folder}")
    if config_folder.exists():
        logging.info(f"Configuration folder: {config_folder} OK")
    else:
        logging.info(f"Configuration folder: {config_folder} will be created")
        config_folder.mkdir()

    # create init-file
    ini_file = config_folder / "luechenbresse.ini"
    config = configparser.ConfigParser()
    if ini_file.exists():
        logging.info(f"Configuration file: {ini_file} OK")
    else:
        logging.info(f"Configuration file: {ini_file} will be created")
        ini_file.touch()

    config.read(ini_file)
    if not config.has_section("databases"):
        config.add_section("databases")

    # create database folder
    if not db_folder:
        db_folder = config_folder
        logging.info("Databases go into default location")
    else:
        db_folder = Path(db_folder).resolve()
        logging.info(f"Databases go into {db_folder}")

    if "folder" in config["databases"]:
        current_db_folder = config["databases"]["folder"]
        logging.info(f"Current folder for databases: {current_db_folder} - will be replaced")
    else:
        logging.info("No database location specified yet")

    logging.info(f"Setting database location to {db_folder}")
    config["databases"]["folder"] = str(db_folder)
    with open(ini_file, "w") as fp:
        config.write(fp)

    if db_folder.exists():
        logging.info("Using existing folder")
    else:
        logging.info(f"Creating {db_folder}...")
        db_folder.mkdir()

    feeds = data.feeds()
    for name in feeds:
        logging.info(f"Installing {name}")
        feed = feeds[name]
        db_file = db_folder / feed["db"]
        schema = feed["schema"]
        if db_file.exists():
            logging.info(f"Database {db_file} exists. LEAVING UNCHANGED")
        else:
            logging.info(f"Database {db_file} will be created with schema {schema}")
            # TODO: buffer schema
            _create_db(db_file, schema)


# TODO move LogManager to module where init() will live
class LogManager(object):
    """
    Set up logging and hold enough context to close the logfile for the current run and send contents via mail
    when all is set and done. A lalzada-Singleton.
    """

    def __new__(cls):
        if not hasattr(cls, 'instance') or not cls.instance:
            FMT = "%(asctime)s [%(levelname)s] %(message)s"
            cls.instance = super().__new__(cls)
            self = cls.instance  # TODO scnr

            # default.log with 10 rotating segments of 100k each -> 1 MB (reicht viele Tage)
            log_file = Path(os.environ["HOME"]) / ".luechenbresse" / "default.log"
            rfh = logging.handlers.RotatingFileHandler(log_file, maxBytes=100_000, backupCount=10)
            rfh.setLevel(logging.INFO)

            # console output TODO switch off via ini-file
            sh = logging.StreamHandler(sys.stdout)
            sh.setLevel(logging.DEBUG)

            # file for output of the current run, will be sent via mail
            self.oneoff_file = Path(os.environ["HOME"]) / ".luechenbresse" / "current.log"
            self.fh = logging.FileHandler(self.oneoff_file, mode="w")
            self.fh.setLevel(logging.DEBUG)

            logging.basicConfig(level=logging.INFO, handlers=[rfh, sh, self.fh], format=FMT)
            logging.info("LogManager lebt.")

        return cls.instance

    def mail(self):
        # close current.log
        # https://stackoverflow.com/questions/15435652/python-does-not-release-filehandles-to-logfile
        logger = logging.getLogger()
        logger.removeHandler(self.fh)
        logging.info(f"closed ~/.luechenbresse/current.log")
        # send file contents via email
        # https://realpython.com/python-pathlib/#reading-and-writing-files
        subject = "von luechenbresse with love"
        body = self.oneoff_file.read_text()
        Mailgun().shoot(subject, body)


if __name__ == "__main__":
    pass


#!/usr/bin/env python
# coding: utf-8

"""
Support setuptools.

Created: 03.05.20
"""

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='luechenbresse',
    version = '0.0.4',
    description = 'Scraping Germany',
    long_description = long_description,
    long_description_content_type = "text/markdown",
    url = 'https://github.com/turkishmaid/luechenbresse',
    author = 'Sara Ziner',
    # author_email = 'flyingcircus@example.com',
    license = 'MIT',
    packages = setuptools.find_packages(),
    include_package_data = True,
    zip_safe = False,
    scripts = [ "bin/luechenbresse" ],
    install_requires = [ 'requests', 'feedparser', 'bs4', 'docopt', ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
)


#!/usr/bin/env python
# coding: utf-8

"""
Convenience API to package data
Following wim's advice in https://stackoverflow.com/a/58941536/3991164

Created: 09.05.20
"""

import logging
import pkgutil
import json as _thejson

_DATA_BUFFER = dict()

def get(name):
    if not name in _DATA_BUFFER:
        logging.info(f'pkgutil.get_data("__name__", "data/{name}")')
        binary = pkgutil.get_data(__name__, f"data/{name}")
        _DATA_BUFFER[name] = binary
    return _DATA_BUFFER[name]

def text(name):
    return get(name).decode("utf-8")

def json(name):
    return _thejson.loads(text(name))

# TODO buffering
def feeds():
    return json("feeds.json")

_SCHEMA_BUFFER = dict()

def schema(name):
    from_buffer = True
    if not name in _SCHEMA_BUFFER:
        from_buffer = False
        sql = text(name)
        _SCHEMA_BUFFER[name] = sql
    return _SCHEMA_BUFFER[name], from_buffer
